# -*- coding: utf8 -*-
import base64
import hashlib
import json
import logging
import os
import random
import socket
import sys

import click
import requests
from pandas.io.json import json_normalize
from requests import HTTPError
from terminaltables import PorcelainTable
from six.moves.urllib import parse
from .utilities.tables import format_json_data, dict_to_csv


def print_as_json(result):
    click.echo(json.dumps(result))


def print_as_bad_json(result):
    click.echo(result)


def create_validate_length(min_length, max_length):
    def validate_length(ctx, option, value):
        if value is None:
            return value

        if len(value) > max_length:
            raise click.BadParameter("needs to be shorter than %s characters" % max_length)

        if len(value) < min_length:
            raise click.BadParameter("needs to be longer than %s characters" % min_length)

        return value

    return validate_length


def build_auth0_url(auth0):
    return '{}.auth0.com'.format(auth0)


def add_to_data_if_not_none(data, val, key):
    if val is not None:
        data[key] = val


def get_data_and_remove(kwargs, key):
    key = key.lower()
    result = kwargs.get(key)

    if key in kwargs:
        del kwargs[key]

    return result


def base64url(b):
    return bytearray(base64.b64encode(b).decode('ascii').replace('=', '').replace('+', '-').replace('/', '_'), 'ascii')


def sha256(s):
    h = hashlib.sha256()
    h.update(s)
    return h.digest()


def urljoin(*args):
    base = args[0]
    for u in args[1:]:
        base = parse.urljoin(base, u)

    return base


class LocalWebServer(object):
    ports = [27017, 41589, 55130, 6617, 1566, 17404, 44288, 6948, 15950, 48216, 3318, 34449, 22082, 22845, 37862, 59304]

    def __init__(self):
        self.app = self.create_app()
        self.code = None
        self.srv = None

        self.disable_logging()
        self.start_server()

    def close(self):
        if self.srv is not None:
            self.srv.server_close()

    @classmethod
    def disable_logging(cls):
        log = logging.getLogger('werkzeug')
        log.setLevel(logging.ERROR)

    def create_app(self):
        from flask import Flask, request, redirect
        app = Flask(__name__)

        @app.route('/console/auth/init')
        def auth_init():
            self.code = request.args.get('code')
            return redirect("/console/auth/finish")

        @app.route('/console/auth/finish')
        def auth_finish():
            self.srv.shutdown_signal = True
            return "You can go now back to the console"

        return app

    def run(self):
        self.srv.serve_forever()

    @property
    def redirect_uri(self):
        return 'http://localhost:%s/console/auth/init' % self.srv.port

    def start_server(self):
        from werkzeug.serving import make_server

        ports = self.ports[:]
        random.shuffle(ports)

        for port in ports:
            try:
                self.srv = make_server('127.0.0.1', port, self.app)
                break
            except (OSError, socket.error):
                continue

        if self.srv is None:
            click.echo("can't create local server for authentication, consider using the --disable-webserver flag")


def wait_for_input():
    if sys.version_info >= (3, 0):
        code = input('Enter the token ')
    else:
        # noinspection PyUnresolvedReferences
        code = raw_input('Enter the token ')

    return code


def auth0_oauth_token(auth0_url, code, verifier, client_id, redirect_uri):
    params = {
        'code': code,
        'code_verifier': verifier,
        'client_id': client_id,
        'grant_type': 'authorization_code',
        'redirect_uri': redirect_uri,
    }

    r = requests.post('https://{}/oauth/token'.format(auth0_url), json=params)

    r.raise_for_status()

    return r.json()


# noinspection PyUnusedLocal
def pixy_flow(ctx):
    import webbrowser

    verifier = base64url(os.urandom(32))
    verifier_challenge = base64url(sha256(verifier))
    verifier_challenge = verifier_challenge.decode('ascii')

    verifier = verifier.decode('ascii')

    def get_authorize_url():
        host = ctx.host if 'missinglink.ai' in ctx.host else '{host}.firebaseapp.com'.format(host=ctx.host)

        current_redirect_uri = app.redirect_uri if ctx.local_web_server else '{host}/console/auth/init'.format(
            host=host)

        query = {
            'response_type': 'code',
            'scope': 'openid offline_access user_external_id org orgs email picture name given_name user_metadata',
            'client_id': ctx.client_id,
            'redirect_uri': current_redirect_uri,
            'code_challenge': verifier_challenge,
            'code_challenge_method': 'S256'
        }

        return 'https://{}/authorize?{}'.format(build_auth0_url(ctx.auth0), parse.urlencode(query)), current_redirect_uri

    app = LocalWebServer() if ctx.local_web_server else None

    authorize_url, redirect_uri = get_authorize_url()

    if ctx.local_web_server and not webbrowser.open(authorize_url):
        app.close()
        ctx.local_web_server = False
        authorize_url, redirect_uri = get_authorize_url()

    if ctx.local_web_server:
        msg = "If the browser doesn't open enter this URL manually\n{authorize_url}\n"
    else:
        msg = "Enter the following URL in your machine browser\n{authorize_url}\n"

    click.echo(msg.format(authorize_url=authorize_url))

    if ctx.local_web_server:
        app.run()
        code = app.code
    else:
        code = wait_for_input()

    try:
        data = auth0_oauth_token(build_auth0_url(ctx.auth0), code, verifier, ctx.client_id, redirect_uri)
    except HTTPError:
        click.echo("Failed to init authorized (did you enter the current token?).", err=True)
        exit(1)
        return

    click.echo("Success!, you sre authorized to use the command line.")

    return data['access_token'], data['refresh_token'], data['id_token']


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


def output_result(ctx, result, fields=None, formatters=None, write_row_index=False):
    if result is None:
        return

    format_tables = ctx.obj.output_format == 'tables'
    format_csv = ctx.obj.output_format == 'csv'

    result = format_json_data(result, formatters)

    def normal_name(name):
        return name.replace('_', ' ').title()

    if format_csv:
        df = json_normalize(result)
        df = df.rename(normal_name, axis='columns')

        df.to_csv(sys.stdout, columns=map(normal_name, fields), index=write_row_index)
    elif format_tables:
        table_data = list(dict_to_csv(result, fields))

        click.echo(PorcelainTable(table_data).table)
    elif hasattr(ctx.obj, 'bad_json_format'):
        print_as_bad_json(result)
    else:
        print_as_json(result)



# -*- coding: utf8 -*-
import os

from .eprint import eprint
from .path_utils import create_dir

try:
    # noinspection PyPep8Naming
    import ConfigParser as configparser
except ImportError:
    # noinspection PyUnresolvedReferences
    import configparser

_missing_link_config = 'missinglink.cfg'


default_api_host = 'https://missinglinkai.appspot.com'
default_host = 'https://missinglink.ai'
default_client_id = 'nbkyPAMoxj5tNzpP07vyrrsVZnhKYhMj'
default_auth0 = 'missinglink'


def get_prefix_section(config_prefix, section):
    return '%s-%s' % (config_prefix, section) if config_prefix else section


def default_missing_link_folder():
    return os.path.join(os.path.expanduser('~'), '.MissingLinkAI')


def default_config_file(config_prefix, filename):
    filename_with_prefix = '%s-%s' % (config_prefix, filename) if config_prefix else filename
    return os.path.join(default_missing_link_folder(), filename_with_prefix)


def find_first_file(config_prefix, filename):
    filename_with_prefix = '%s-%s' % (config_prefix, filename) if config_prefix else filename

    possible_paths = [os.getcwd(), default_missing_link_folder()]

    for possible_path in possible_paths:
        candidate_config_file = os.path.join(possible_path, filename_with_prefix)

        if os.path.isfile(candidate_config_file):
            return candidate_config_file

    return None


class Config(object):
    def __init__(self, config_prefix=None, config_file=None, **kwargs):
        self.config_file = config_file
        self.config_prefix = config_prefix

        if self.config_file:
            self.config_file_abs_path = os.path.abspath(config_file)
        else:
            self.config_file_abs_path = find_first_file(self.config_prefix, _missing_link_config)

        if config_file and not os.path.exists(self.config_file_abs_path):
            eprint('config file %s not found' % self.config_file_abs_path)
            exit(1)
        elif config_prefix and self.config_file_abs_path is None:
            eprint('config file for %s not found' % config_prefix)
            exit(1)

        self.config_prefix = config_prefix

        parser = configparser.RawConfigParser()
        readonly_parser = configparser.RawConfigParser()

        if self.config_file_abs_path is not None:
            parser.read([self.config_file_abs_path])
            readonly_parser.read([self.config_file_abs_path])

        self.parser = parser
        self.readonly_parser = readonly_parser

        self.__append_kwargs(**kwargs)

    def __append_kwargs(self, **kwargs):
        for section, values in kwargs.items():
            try:
                self.parser.add_section(section)
                self.readonly_parser.add_section(section)
            except configparser.DuplicateSectionError:
                pass

            for key, val in values.items():
                self.parser.set(section, key, val)
                self.readonly_parser.set(section, key, val)

    @property
    def init_dict(self):
        return dict(config_file=self.config_file, config_prefix=self.config_prefix)

    @property
    def api_host(self):
        return self.general_config.get('api_host', default_api_host)

    @property
    def host(self):
        host = self.general_config.get('host', default_host)

        if not host.startswith('http://') and not host.startswith('https://'):
            host = 'https://' + host

        return host

    @property
    def client_id(self):
        return self.general_config.get('client_id', default_client_id)

    @property
    def auth0(self):
        return self.general_config.get('auth0', default_auth0)

    def get_prefix_section(self, section):
        return get_prefix_section(self.config_prefix, section)

    @property
    def token_config(self):
        section_name = self.get_prefix_section('token')
        return self.items(section_name)

    @property
    def general_config(self):
        return self.readonly_items('general', most_exists=self.config_prefix is not None)

    @property
    def refresh_token(self):
        return self.token_config.get('refresh_token')

    @property
    def id_token(self):
        return self.token_config.get('id_token')

    @property
    def token_data(self):
        import jwt

        return jwt.decode(self.id_token, verify=False) if self.id_token else {}

    @property
    def user_id(self):
        import jwt

        data = jwt.decode(self.id_token, verify=False) if self.id_token else {}

        return data.get('user_external_id')

    @classmethod
    def items_from_parse(cls, parser, section, most_exists):
        try:
            return dict(parser.items(section))
        except configparser.NoSectionError:
            if most_exists:
                raise

            return {}

    def items(self, section, most_exists=False):
        return self.items_from_parse(self.parser, section, most_exists=most_exists)

    def readonly_items(self, section, most_exists=False):
        return self.items_from_parse(self.readonly_parser, section, most_exists=most_exists)

    def set(self, section, key, val):
        try:
            self.parser.add_section(section)
        except configparser.DuplicateSectionError:
            pass

        self.parser.set(section, key, val)

    def _write(self, fo):
        self.parser.write(fo)

    def save(self):
        # we always save the config prefix file into the default folder
        if self.config_file_abs_path is None:
            self.config_file_abs_path = find_first_file(self.config_prefix, _missing_link_config) if not self.config_prefix else None
            self.config_file_abs_path = self.config_file_abs_path or default_config_file(self.config_prefix, _missing_link_config)

        create_dir(os.path.dirname(self.config_file_abs_path))

        with open(self.config_file_abs_path, 'w') as configfile:
            self._write(configfile)

    def update_and_save(self, d):
        for section, section_values in d.items():
            section_name_with_prefix = self.get_prefix_section(section)
            for key, val in section_values.items():
                self.set(section_name_with_prefix, key, val)

        self.save()



# -*- coding: utf8 -*-
import json
import logging
import tempfile
from contextlib import closing
import sys
import os
from uuid import uuid4
import datetime
import requests
import six
from six import wraps

from .api import default_api_retry
from .utils.os_utils import open_atomic
from .config import Config
from .data_volume import with_repo
from .json_utils import multi_line_json_from_data, normalize_item, dict_normalize
from . import MetadataOperationError
from .multi_process_control import get_multi_process_control
from .path_utils import get_batch_of_files_from_paths, safe_make_dirs, DestPathEnum, enumerate_paths_with_info
from .print_status import PrintStatus
from .eprint import eprint


def wrap_exceptions(fn):
    @wraps(fn)
    def decorated(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except AssertionError:
            raise
        except Exception as ex:
            eprint('\n' + str(ex))
            sys.exit(1)

    return decorated


epoch = datetime.datetime.utcfromtimestamp(0)


class InvalidJsonFile(Exception):
    def __init__(self, filename):
        self.filename = filename


def _download_entity(config_init_dict, volume_id, dest_file, item_id, item_meta, save_meta=False):
    DataSync.download_entity(config_init_dict, volume_id, dest_file, item_id, item_meta, save_meta=save_meta)


def chop_microseconds(delta):
    return delta - datetime.timedelta(microseconds=delta.microseconds)


class DataSync(object):
    metadata_ext = '.metadata.json'

    def __init__(self, ctx, repo, no_progressbar):
        self.__ctx = ctx
        self.__repo = repo
        self.__no_progressbar = no_progressbar

    @property
    def repo(self):
        return self.__repo

    def __upload_file_for_processing(self, file_obj, file_description):
        from .gcs_utils import do_upload
        from tqdm import tqdm

        volume_id = self.__repo.data_volume_config.volume_id

        data_object_name = '%s/temp/%s_%s.json' % (volume_id, file_description, uuid4().hex)

        url = 'data_volumes/{volume_id}/gcs_urls'.format(volume_id=volume_id)

        headers = {'Content-Type': 'application/json'}

        msg = {
            'methods': 'PUT',
            'paths': [data_object_name],
            'content_type': 'application/json',
            'temp': True,
        }

        result = self.__ctx.obj.handle_api(self.__ctx.obj, requests.post, url, msg, retry=default_api_retry())

        put_url = result['put'][0]

        gcs_auth = self.__repo.data_volume_config.object_store_config.get('auth')

        def update_bar(c):
            bar.update(c)

        logging.debug('uploading %s', data_object_name)
        with tqdm(total=file_obj.tell(), desc="Uploading {}".format(file_description), unit=' KB', ncols=80,
                  disable=self.__no_progressbar, unit_scale=True) as bar:
            do_upload(gcs_auth, None, data_object_name, file_obj, headers, head_url=None, put_url=put_url,
                      callback=update_bar)

        return data_object_name

    def __process_index(self, object_name):
        index = self.__repo.open_index()

        change_set = self.__process_with_status('Index', lambda: index.get_changeset(data='gs://' + object_name))

        total_modify_files = 0
        total_new_files = 0

        files_to_upload = tempfile.TemporaryFile('w+')

        for name, op in change_set:
            if op == 'm':
                total_modify_files += 1
            elif op == 'i':
                total_new_files += 1
            else:
                continue

            data = name + '\n'

            files_to_upload.write(data)

        total_files_to_upload = total_modify_files + total_new_files
        files_to_upload.seek(0)

        return files_to_upload, total_files_to_upload

    @classmethod
    def __process_with_status(cls, process_type, callback):
        start_time = datetime.datetime.utcnow()
        status_line = PrintStatus()
        with closing(status_line):
            status_line.print_status('Processing %s' % process_type)
            result = callback()
            total_time = datetime.datetime.utcnow() - start_time
            total_time = chop_microseconds(total_time)
            status_line.print_status('Processing %s Done (%s)' % (process_type, total_time))

        return result

    def upload_and_update_index(self, index_entities_file):
        object_name = self.__upload_file_for_processing(index_entities_file, 'Index')

        files_to_upload, total_files_to_upload = self.__process_index(object_name)

        return files_to_upload, total_files_to_upload

    def upload_and_update_metadata(self, file_obj):
        object_name = self.__upload_file_for_processing(file_obj, 'Metadata')

        self.__process_with_status('Metadata', lambda: self.__repo.metadata.add_data(data='gs://' + object_name))

    @classmethod
    def __meta_companion_file_name(cls, metadata_file_name):
        return metadata_file_name[:-len(cls.metadata_ext)]

    def __get_metadata_info(self, rel_metadata_file_name):
        full_path_metadata = os.path.join(self.__repo.data_path, rel_metadata_file_name)

        rel_path_key = rel_metadata_file_name[:-len(self.metadata_ext)]

        with open(full_path_metadata) as metadata_file:
            try:
                data = json.load(metadata_file)
                for key, val in data.items():
                    if isinstance(val, (dict, list)):
                        data[key] = json.dumps(val)
            except ValueError:
                raise InvalidJsonFile(rel_metadata_file_name)

            return rel_path_key, data

    def __append_index_file(self, file_info, combined_index_files):
        file_name = file_info['path']

        mtime = file_info['mtime']
        ctime = file_info.get('ctime', mtime)

        mtime = (mtime - epoch).total_seconds()
        ctime = (ctime - epoch).total_seconds()

        if file_info['sys'] == 'local':
            rel_path = os.path.relpath(file_name, self.__repo.data_path)

            params = {
                'ctime': ctime,
                'mtime': mtime,
                'size': file_info['size'],
                'sha': file_info['sha'],
                'mode': 0,
            }
        else:  # file_info['sys'] == 's3':
            rel_path = file_name

            params = {
                'ctime': mtime,
                'mtime': ctime,
                'size': file_info['size'],
                'sha': file_info['sha'],
                'mode': 0,
            }

        multi_line_json_from_data({rel_path: params}, combined_index_files)

    def __load_all_metadata(self, data_files_info, metadata_files_list):
        from tqdm import tqdm

        files_metadata = {}
        for rel_file_name in tqdm(metadata_files_list, desc="Read metadata", unit=' files', ncols=80, disable=self.__no_progressbar):
            if self.__meta_companion_file_name(rel_file_name) in data_files_info:
                key, val = self.__get_metadata_info(rel_file_name)

                files_metadata[key] = val
            else:
                logging.debug("file %s doesn't have a data point file", rel_file_name)

        return files_metadata

    def create_combined_index_and_metadata(self, data_path):
        import humanize

        total_files = 0
        total_files_size = 0
        total_metadata = 0
        status_line = PrintStatus()

        data_files_list = set()
        metadata_files_list = set()
        data_files_info = {}

        with closing(status_line):
            for file_info in enumerate_paths_with_info(data_path):
                rel_file_name = self.repo.rel_path(file_info['path'])
                total_files_size += file_info['size']

                if rel_file_name.endswith(self.metadata_ext):
                    total_metadata += 1
                    metadata_files_list.add(rel_file_name)
                else:
                    total_files += 1
                    data_files_list.add(rel_file_name)
                    data_files_info[rel_file_name] = file_info

                status_line.print_status(
                    'Total files {:,} ({}) (metadata found: {:,})', total_files,
                    humanize.naturalsize(total_files_size), total_metadata)

        files_metadata = self.__load_all_metadata(data_files_info, metadata_files_list)

        return files_metadata, data_files_info

    @wrap_exceptions
    def upload_in_batches(self, files_info, total_files=None, callback=None):
        total_files = total_files or len(files_info)
        batch_size = max(min(total_files // 100, 250), 250)  # FIXME: hardcoded

        for files_info_batch in get_batch_of_files_from_paths(files_info, batch_size):
            self.__repo.stage(files_info_batch, callback=callback)

    @classmethod
    def download_entity(cls, config_init_dict, volume_id, dest_file, item_id, item_meta, save_meta=True):
        import google.api_core.exceptions

        config = Config(**config_init_dict)

        with with_repo(config, volume_id, read_only=True) as repo:
            safe_make_dirs(os.path.dirname(dest_file))

            if not os.path.exists(dest_file):
                try:
                    _, data = repo.object_store.get_raw(item_id)
                except requests.exceptions.HTTPError as ex:
                    if ex.response.status_code == 404:
                        return

                    raise
                except google.api_core.exceptions.NotFound:
                    return

                with open_atomic(dest_file, 'wb') as f:
                    f.write(data)

            if save_meta:
                json_metadata_name = dest_file + '.metadata.json'

                with open(json_metadata_name, 'w') as metadata_file:
                    json.dump(item_meta, metadata_file)

    def download_all(self, query, root_folder, dest_pattern, batch_size, processes, save_meta=True):
        multi_process_control = get_multi_process_control(processes)
        return self.__download_all(multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=save_meta)

    def __download_results(self, process_control, dest_pattern, results, callback=None, save_meta=False):
        volume_id = self.__repo.data_volume_config.volume_id

        def wrap_callback_func(current_item, current_full_path):
            def wrapper(_):
                if callback is not None:
                    callback(current_full_path, current_item)

            return wrapper

        for item in results:
            normalized_item = normalize_item(item)
            config_init_dict = self.__ctx.obj.config.init_dict

            full_path = DestPathEnum.get_full_path(dest_pattern, normalized_item)

            yield process_control.execute(
                _download_entity, args=(config_init_dict, volume_id, full_path, item['id'], item.get('meta', {}), save_meta),
                callback=wrap_callback_func(item, full_path)), full_path, normalized_item

    class DownloadIterResults(object):
        def __init__(self, repo, query, batch_size):
            self.__query = query
            self.__batch_size = batch_size
            self.__repo = repo
            self.__query_index = 0
            self.results = None
            self.total_data_points = None

        def __get_next_results(self):
            current_results, current_total_data_points = self.__repo.metadata.query(
                self.__query, max_results=self.__batch_size, start_index=self.__query_index)

            return list(current_results), current_total_data_points

        def skip(self, page_number):
            self.__query_index = page_number * self.__batch_size

        def next(self):
            results, total_data_points = self.__get_next_results()

            self.results = results
            self.total_data_points = total_data_points

            self.__query_index += len(results)

            return results, total_data_points

    def begin_download_iter(self, query, batch_size):
        return self.DownloadIterResults(self.__repo, query, batch_size)

    def download_iter(self, iter_results, dest_pattern, processes=-1):
        multi_process_control = get_multi_process_control(processes)

        with closing(multi_process_control):
            try:
                results, _ = next(iter_results)

                if len(results) == 0:
                    return

                futures = []
                for async_result, full_path, metadata in self.__download_results(multi_process_control, dest_pattern, results):
                    futures.append((async_result, full_path, metadata))

                for async_result, full_path, metadata in futures:
                    async_result.wait()

                    yield full_path, metadata
            except StopIteration:
                pass
            except MetadataOperationError as ex:
                eprint(str(ex))
            except KeyboardInterrupt:
                multi_process_control.terminate()
                multi_process_control.close()

    def __download_all(self, multi_process_control, query, root_folder, dest_pattern, batch_size, save_meta=True):
        from tqdm import tqdm

        def get_next_results(query_index):
            current_results, current_total_data_points = self.__repo.metadata.query(
                query, max_results=batch_size, start_index=query_index)

            return list(current_results), current_total_data_points

        def handle_item(file_name, item):
            rel_path = os.path.relpath(file_name, root_folder)
            phase = item.get('phase', 'all')
            phase_meta.setdefault(phase, {})[rel_path] = item

            bar.update()

        start_index = 0

        results, total_data_points = get_next_results(start_index)

        phase_meta = {}

        with tqdm(total=total_data_points, desc="Downloading", unit=' data point', ncols=80, disable=self.__no_progressbar) as bar:
            with closing(multi_process_control):
                try:
                    while True:
                        for _ in self.__download_results(multi_process_control, dest_pattern, results, handle_item, save_meta=save_meta):
                            pass

                        if len(results) != batch_size:
                            break

                        start_index += len(results)

                        results, _ = get_next_results(start_index)
                except MetadataOperationError as ex:
                    eprint(str(ex))
                except KeyboardInterrupt:
                    multi_process_control.terminate()
                    multi_process_control.close()

        return phase_meta

    def save_metadata(self, root_dest, metadata):
        from tqdm import tqdm

        with tqdm(total=len(metadata), desc="saving metadata", unit=' files', ncols=80, disable=self.__no_progressbar) as bar:
            for key, val in metadata.items():
                if key is None:
                    key = 'unknown'

                json_metadata_file = os.path.join(root_dest, key + '.metadata.json')

                with open(json_metadata_file, 'w') as f:
                    json.dump(val, f)

                bar.update()

    def create_index_file_from_data(self, f, files_info):
        for val in files_info.values():
            self.__append_index_file(val, f)

    @classmethod
    def create_metadata_file_from_data(cls, f, metadata_files):
        dict_normalize(metadata_files)

        for rel_file_name, file_metadata in metadata_files.items():
            multi_line_json_from_data({rel_file_name: file_metadata}, f)

    def upload_index_and_metadata(self, data_path):
        metadata_files, files_info = self.create_combined_index_and_metadata(data_path)

        if len(files_info) == 0:
            return []

        if len(metadata_files) > 0:
            combined_meta_files = tempfile.TemporaryFile('wb+')
            self.create_metadata_file_from_data(combined_meta_files, metadata_files)
            self.upload_and_update_metadata(combined_meta_files)

        combined_index_files = tempfile.TemporaryFile('wb+')
        self.create_index_file_from_data(combined_index_files, files_info)
        files_to_upload, total_files_to_upload = self.upload_and_update_index(combined_index_files)

        def switch_to_full_path(params):
            rel_path = params['path']

            params['path'] = self.repo.full_path(rel_path)

            return params

        def enum_file_names():
            for file_name in files_to_upload:
                file_name = file_name.strip()

                yield file_name

        files_to_upload = [switch_to_full_path(files_info[rel_file_name]) for rel_file_name in enum_file_names()]

        return files_to_upload



# -*- coding: utf8 -*-
import logging
import os
import os.path
import shutil


def should_update(src_file, dest_file):
    try:
        sink_st = os.stat(src_file)
    except FileNotFoundError:
        logging.error("Fail to retrieve information about sink %s (skip update)", src_file)
        return False

    sink_sz = sink_st.st_size
    sink_mt = sink_st.st_mtime

    try:
        target_st = os.stat(dest_file)
    except FileNotFoundError:
        logging.error("Fail to retrieve information about target %s (skip update)", dest_file)
        return False

    target_sz = target_st.st_size
    target_mt = target_st.st_mtime

    if target_sz != sink_sz:
        return True

    return target_mt != sink_mt


def copy_file(src_file, dest_file):
    logging.debug("copy: %s to: %s", src_file, dest_file)

    try:
        shutil.copyfile(src_file, dest_file)
    except IOError:
        logging.error("Fail to copy %s", src_file)

    try:
        s = os.stat(src_file)
        os.utime(dest_file, (s.st_atime, s.st_mtime))
    except IOError:
        logging.error("Fail to copy timestamp of %s", src_file)


def copy_from_to(src_root, src_files, dest):
    for src_name in src_files:
        rel_path = os.path.relpath(src_name, src_root)

        dest_file = os.path.join(dest, rel_path)

        if os.path.isfile(dest_file) and not should_update(src_name, dest_file):
            continue

        dest_dir_name = os.path.dirname(dest_file)

        if not os.path.isdir(dest_dir_name):
            os.makedirs(dest_dir_name, exist_ok=True)

        copy_file(src_name, dest_file)



# -*- coding: utf8 -*-
import boto3
import botocore
from hashlib import sha1

import os

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


def _hex_to_filename(path, hex):
    """Takes a hex sha and returns its filename relative to the given path."""
    # os.path.join accepts bytes or unicode, but all args must be of the same
    # type. Make sure that hex which is expected to be bytes, is the same type
    # as path.
    if getattr(hex, 'decode', None) is not None:
        hex = hex.decode('ascii')

    if getattr(path, 'decode', None) is not None:
        path = path.decode('ascii')

    dir_name = hex[:2]
    file = hex[2:]
    # Check from object dir
    return os.path.join(path, dir_name, file)


def transfer_object(event, context):
    src_bucket = event['src_bucket']
    dest_bucket = event['dest_bucket']
    src_object = event['src_object']

    s3 = boto3.resource('s3')

    with StringIO() as f:
        s3.Bucket(src_bucket).download_fileobj(src_object, f)

        f.seek(0)

        h = sha1()
        h.update(f.getvalue())
        sha = h.hexdigest()

        dest_object = _hex_to_filename('objects', sha)

        s3.Bucket(dest_bucket).upload_fileobj(src_object, dest_object)

        return {
            'sha': sha
        }



# -*- coding: utf8 -*-
from __future__ import print_function
import sys
from .update_timer import UpdateTimer


class PrintStatus(object):
    def __init__(self, nl_on_clone=True):
        self.__update_timer = UpdateTimer()
        self.__first_status = True
        self.__last_printed_msg = None
        self.__last_msg = None
        self.__nl_on_close = nl_on_clone

    def close(self):
        if self.__last_msg is not None and self.__last_msg != self.__last_printed_msg:
            self.__print_status(True, self.__last_msg)

        if self.__nl_on_close:
            self.__print_to_stdout('\n')

    def __get_formatted_message(self, msg, *args, **kwargs):
        formatted_msg = ''

        if self.__last_printed_msg:
            spaces = len(self.__last_printed_msg)
            formatted_msg += '\r' + (' ' * spaces) + '\r'

        formatted_msg += msg.format(*args, **kwargs)

        if self.__first_status:
            formatted_msg = '\n' + formatted_msg

        return formatted_msg

    @classmethod
    def __print_to_stdout(cls, text):
        print(text, end='')
        sys.stdout.flush()

    def __print_status(self, force, msg, *args, **kwargs):
        formatted_msg = self.__get_formatted_message(msg, *args, **kwargs)

        self.__last_msg = formatted_msg.strip()

        if self.__update_timer.can_update() or force:
            self.__print_to_stdout(formatted_msg)
            self.__last_printed_msg = self.__last_msg

        return True

    def print_status(self, msg, *args, **kwargs):
        printed_status = self.__print_status(False, msg, *args, **kwargs)

        if self.__first_status and printed_status:
            self.__first_status = False



# -*- coding: utf8 -*-
import os
import errno
import requests
from contextlib import closing
from ..dulwich.errors import NotGitRepository
from ..data_volume_config import DataVolumeConfig
from ..path_utils import expend_and_validate_path
from .repo import MlRepo


def default_data_volume_path(volume_id):
    return os.path.join(os.path.expanduser('~'), '.MissingLinkAI', 'data_volumes', str(volume_id))


def update_config_file(volume_id, params):
    path = default_data_volume_path(volume_id)

    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    config = DataVolumeConfig(path)

    config.update_and_save(params)

    return config


def create_data_volume(volume_id, data_path, linked, display_name, description, **kwargs):
    params = {
        'general': {
            'id': volume_id,
            'embedded': not linked,
            'datapath': data_path,
            'display_name': display_name,
            'description': description
        }
    }

    params.update(**kwargs)
    return update_config_file(volume_id, params)


def map_volume(ctx, volume_id, data_path):
    result = ctx.obj.handle_api(ctx.obj, requests.get, "data_volumes/{volume_id}".format(volume_id=volume_id))

    if data_path is not None:
        data_path = expend_and_validate_path(data_path)

    bucket_name = result.get('bucket_name')

    params = {}
    if bucket_name is not None:
        params['object_store'] = {'bucket_name': bucket_name}

    config = create_data_volume(
        result['id'],
        data_path,
        not result['embedded'],
        result.get('display_name'),
        result.get('description'),
        **params)

    if bucket_name is None:
        config.remove_option('object_store', 'bucket_name')
        config.save()

    return config


def with_repo_dynamic(ctx, volume_id, data_path=None):
    repo_root = default_data_volume_path(volume_id)

    config = ctx.obj.config

    try:
        repo = MlRepo(config, repo_root)
        if data_path is not None:
            repo.data_path = data_path

        return closing(repo)
    except NotGitRepository:
        data_volume_config = map_volume(ctx, volume_id, data_path)
        return closing(MlRepo(
            config, repo_root, read_only=True, require_path=False, data_volume_config=data_volume_config))


def create_repo(config, repo_root, read_only=False, require_map=True):
    return MlRepo(config, repo_root, read_only=read_only, require_path=require_map)


def with_repo(config, volume_id, read_only=False, require_map=True):
    repo_root = default_data_volume_path(volume_id)

    try:
        return closing(create_repo(config, repo_root, read_only=read_only, require_map=require_map))
    except NotGitRepository:
        msg = 'Data volume {0} was not mapped on this machine, you should call "mali data map {0} --dataPath [root path of data]" in order to work with the volume locally'.format(volume_id)
        click.echo(msg, err=True)
        exit(1)



# -*- coding: utf8 -*-
import logging
import multiprocessing
import uuid
from multiprocessing import Pool, Semaphore
import os
import signal


def process_worker_init(lock_, parent_id):
    def sig_int(signal_num, frame):
        os.kill(parent_id, signal.SIGSTOP)

    signal.signal(signal.SIGINT, sig_int)
    global process_controller_lock
    process_controller_lock = lock_


class MethodWrapper(object):
    def __init__(self, callable_func, lock):
        self.__callable = callable_func
        self.__lock = lock

    def __call__(self, *args, **kwargs):
        lock = self.__lock or process_controller_lock
        with lock:
            try:
                self.__callable(*args, **kwargs)
            except KeyboardInterrupt:
                pass


class _MultiProcessControlShim(object):
    @classmethod
    def execute(cls, method, args, callback=None):
        method(*args)
        if callback:
            callback(None)

    def terminate(self):
        pass

    def close(self):
        pass

    def join(self):
        pass


def get_multi_process_control(processes, use_threads=False):
    if processes == 1:
        return _MultiProcessControlShim()

    return _MultiProcessControl(processes, use_threads=use_threads)


class _MultiProcessControl(object):
    def __use_threads(self, processes):
        from multiprocessing.pool import ThreadPool

        self.__using_threads = True
        self.__processing_pool = ThreadPool(processes)

    def __init__(self, processes=-1, max_pending=-1, use_threads=False):
        processes = multiprocessing.cpu_count() * 2 if processes == -1 else processes
        max_pending_processes = processes * 10 if max_pending == -1 else max_pending
        self.__max_waiting_semaphore = Semaphore(max_pending_processes)
        self.__using_threads = False

        if use_threads:
            self.__use_threads(processes)
        else:
            try:
                self.__processing_pool = Pool(processes, process_worker_init, initargs=(self.__max_waiting_semaphore, os.getpid(),))
            except AssertionError:
                self.__use_threads(processes)

        self.__jobs = {}

    def join(self):
        logging.debug('%s pool joining', self.__class__)
        self.__wait_pending_jobs()
        self.__check_pending_jobs()
        logging.debug('%s pool joined', self.__class__)

    def close(self):
        if self.__processing_pool is not None:
            logging.debug('%s closing & joining pool', self.__class__)
            self.__processing_pool.close()
            self.join()

    def __wait_pending_jobs(self):
        for token, async_result in self.__jobs.items():
            if async_result is not None:
                async_result.wait()

        self.__check_pending_jobs()

    def __check_pending_jobs(self):
        for token, async_result in self.__jobs.items():
            if async_result is not None:
                if not async_result.ready():
                    continue

                async_result.get()
                self.__jobs[token] = None

        self.__jobs = {token: async_result for token, async_result in self.__jobs.items() if async_result is not None}

    def execute(self, method, args, callback=None):
        self.__check_pending_jobs()

        if self.__using_threads:
            method_wrapper_args = (method, self.__max_waiting_semaphore)
        else:
            method_wrapper_args = (method, None)

        token = uuid.uuid4()
        job_async_result = self.__processing_pool.apply_async(
            MethodWrapper(*method_wrapper_args),
            args=args,
            callback=callback)

        self.__jobs[token] = job_async_result

        return job_async_result

    def terminate(self):
        self.__processing_pool.terminate()
        self.__jobs = {}



# -*- coding: utf8 -*-
import click
import requests
import json
from mali_commands.commons import output_result
from mali_commands.options import project_id_option, experiment_id_option, metrics_option, model_weights_hash_option
from mali_commands.options import chart_x_option, chart_x_name_option, chart_y_name_option, chart_y_option, chart_name_option, chart_type_option, \
    chart_scope_option
from mali_commands.utilities.formatters import format_datetime, truncate_long_text, truncate_short_text
import six


def chart_name_to_id(chart_name):
    from hashlib import sha256
    return sha256(six.text_type(chart_name).encode('utf-8').lower().strip()).hexdigest()


@click.group('experiments')
def experiments_commands():
    pass


@experiments_commands.command('list')
@project_id_option(required=True)
@click.pass_context
def list_experiments(ctx, projectid):
    """List experiments of a project.
    """

    experiments = []

    def remove_prefix(name, prefix):
        if name.startswith(prefix):
            return name[:len(name)]

        return name

    def get_fields_names():
        fields = set()

        for experiment in new_experiments:
            if 'values' not in experiment:
                continue

            for value in experiment['values']:
                name = value['key']

                name = remove_prefix(name, 'ml_train_')
                name = remove_prefix(name, 'ml_')

                fields.add(name)

                experiment[name] = value['val']

            del experiment['values']

        displayed_fields.extend(fields)

    displayed_fields = ['experiment_id', 'created_at', 'display_name', 'description']

    cursor_token = None
    while True:
        list_experiments_path = 'projects/{project_id}/experiments'.format(project_id=projectid)

        if cursor_token:
            list_experiments_path += '?cursor_token={cursor_token}'.format(cursor_token=cursor_token)

        result = ctx.obj.handle_api(ctx.obj, requests.get, list_experiments_path)
        cursor_token = result.get('next_experiments_token', '')
        new_experiments = result.get('experiments', [])
        experiments.extend(new_experiments)

        get_fields_names()

        if len(new_experiments) == 100:
            continue

        break

    formatters = {
        'created_at': format_datetime,
        'display_name': truncate_short_text,
        'description': truncate_long_text,
    }
    output_result(ctx, experiments, displayed_fields, formatters=formatters)


@experiments_commands.command('updateMetrics')
@project_id_option(required=False)
@experiment_id_option(required=False)
@model_weights_hash_option(required=False)
@metrics_option(required=True)
@click.pass_context
def update_metrics(ctx, projectid, experimentid, weightshash, metrics):
    """Send external metrics in the experiment level.

    There are 2 ways to identify the experiment: (1) specify both `--projectId` and `--experimentId`
    options or (2) specify `--weightsHash` option. When the model's weights hash is specified, mali
    would look up the experiment that the model belongs to.

    When both `--projectId` and `--experimentId` are specified, the `--weightsHash` option is ignored.

    Example:

    To send metrics to the 5th experiment of the project "123", run

    \b
        mali experiments update_metrics --projectId 123 --experimentId 5 --metrics '{"ex_cost": 99}'

    Or assuming that the model's weights hash "324e16b5e" was generated by this experiment at certain
    epoch or iteration, run

    \b
        mali experiments update_metrics --weightsHash 324e16b5e --metrics '{"ex_cost": 99}'
    """

    update_metrics_path = get_submit_path('projects/{project_id}/experiments/{experiment_id}/metrics', 'model_weights_hashes/{model_weights_hash}/metrics?experiment_only=1', projectid, experimentid, weightshash)
    data = _get_metrics_json(metrics)
    result = ctx.obj.handle_api(ctx.obj, requests.post, update_metrics_path, data)
    output_result(ctx, result, ['ok'])


def get_submit_path(by_project_and_experiment_path, by_model_hash_path, project_id, experiment_id, model_weights_hash, **kwargs):
    def raise_data_missing_error():
        if project_id is not None:
            raise click.BadOptionUsage('Please also specify --experimentId option.')
        elif experiment_id is not None:
            raise click.BadOptionUsage('Please also specify --projectId option.')
        else:
            raise click.BadOptionUsage('Please specify the experiment using (1) --projectId and --experimentId options'
                                       'or (2) --weightsHash options.')

    if project_id is not None and experiment_id is not None:
        return by_project_and_experiment_path.format(project_id=project_id, experiment_id=experiment_id, **kwargs)
    elif model_weights_hash is not None:
        return by_model_hash_path.format(model_weights_hash=model_weights_hash, **kwargs)
    raise_data_missing_error()


def _read_norm_y_values(ys):
    res = []
    dimmention_count = None

    for y in ys:
        if not isinstance(y, list):
            y = [y]
        cur_dim = len(y)
        dimmention_count = dimmention_count or cur_dim
        if dimmention_count != cur_dim:
            raise click.BadOptionUsage("All of the data values arrays must be of the same size")
        res += y
    return res


def _read_norm_x_values(xs):
    res = []
    for type, suffix in [(float, '_float'), (six.integer_types, '_int'), (six.string_types, '_str')]:
        for x in xs:
            if not isinstance(x, type):
                res = []
                break
            res.append(x)
        if len(res) == len(xs):
            return res, suffix
    raise click.BadOptionUsage('X values must be consistent')


def _get_metrics_json(metrics_string):
    try:
        return json.loads(metrics_string)
    except ValueError:
        raise click.BadParameter('The supplied sting is not a valid JSON dictionary format.')


@experiments_commands.command('updateChart')
@model_weights_hash_option(required=False)
@project_id_option(required=False)
@experiment_id_option(required=False)
@chart_name_option(required=True)
@chart_x_name_option(required=False)
@chart_y_name_option(required=False)
@chart_scope_option(required=False)
@chart_type_option(required=False)
@chart_x_option(required=False)
@chart_y_option(required=False)
@click.pass_context
def update_chart(ctx, weightshash, projectid, experimentid, chartname, chartscope, charttype, chartxname, chartyname,
                 chartx, charty):
    """Send experiment external chart to an experiment.

    There are 2 ways to identify the experiment: (1) specify both `--projectId` and `--experimentId`
    options or (2) specify `--weightsHash` option. When the model's weights hash is specified, mali
    would look up the experiment that the model belongs to.

    When both `--projectId` and `--experimentId` are specified, the `--weightsHash` option is ignored.

    To send chart to the 5th experiment "precision recall" chart of the project "123" in the validation scope, run

    \b
        mali experiments updateChart --projectId 123 --experimentId 321  --chartName "precision recall" --chartScope validation  --chartX '[0.1,0.5,0.8]' --chartY '[0.9,0.5,0.2]' --chartXNname "Precision" --chartYName "Recall"

    Or send a chart with multiple y charts:
        mali experiments updateChart --p 123 --e 321  -c "precision recall" -cs validation  -cx '[0.1,0.5,0.8]' -cy '[[0.9, 0.5],[0.5,0.25],[0.2,0.4]' --chartYName '["Func1","Func2"]'

    """
    chart_id = chart_name_to_id(chartname)

    update_chart_path = get_submit_path(
        'projects/{project_id}/experiments/{experiment_id}/chart/{chart_id}',
        'model_weights_hashes/{model_weights_hash}/chart/{chart_id}',
        projectid, experimentid, weightshash, chart_id=chart_id)

    x, x_suffix = _read_norm_x_values(_get_metrics_json(chartx))

    y = _read_norm_y_values(_get_metrics_json(charty))
    y_name = chartyname
    if y_name.startswith('['):
        y_name = _get_metrics_json(y_name)
    if not isinstance(y_name, list):
        y_name = [y_name]
    chartlegends = [chartxname] + y_name
    if len(y) % len(x) is not 0:
        raise click.BadOptionUsage('X and Y arrays must be of the same size')
    data = {'name': chartname, 'labels': chartlegends,
            'x' + x_suffix: x, 'y_data': y, 'chart': charttype,
            'scope': chartscope}
    result = ctx.obj.handle_api(ctx.obj, requests.put, update_chart_path, data)
    output_result(ctx, result, ['ok'])


@experiments_commands.command('updateMetricsPerIteration')
@model_weights_hash_option(required=True)
@metrics_option(required=True)
@click.pass_context
def update_model_metrics(ctx, weightshash, metrics):
    """
    Send external metrics for a specific iteration.
    Using this option you can create a graph of the metrics for each iteration.
    The experiment and iterations are obtained from the weighted hash.

    Each model weights hash is attached to certain experiment epochs and thus can be used to send
    metrics that are relevant to those epochs. The model weights hash is a hexadecimal string. To
    calculate the weights hash of a model:

        - Calculate the sha1 strings of the weights for each layers

        - Calculate the sha1 string of the combine hashes.
            For example, the model has 3 layers with the layers' weight hashes: ['abc', '123', 'def'], the model weight hash is `sha1('abc123def')`

    WARNING: The same model weights hash can appear in different experiments or different epochs
    (for example, when running the same net twice). As such, this command will send the metrics to
    all these experiments/epochs that it can identify from the hash.

    Example:

    To send metrics of the model which hash is 324e16b5e, run

    \b
        mali experiments updateMetricsPerIteration --weightsHash 324e16b5e --metrics '{"ex_cost": 99}'

    """
    update_model_metrics_path = 'model_weights_hashes/{model_weights_hash}/metrics' \
        .format(model_weights_hash=weightshash)
    data = _get_metrics_json(metrics)

    result = ctx.obj.handle_api(ctx.obj, requests.post, update_model_metrics_path, data)
    output_result(ctx, result, ['ok'])
